import os
from contextlib import asynccontextmanager
from functools import partial
from pathlib import Path
from typing import Any

from anyio import sleep_forever
from anyio.abc import TaskStatus
from fastapi import Depends, FastAPI, WebSocket, WebSocketDisconnect
from fastapi.middleware.cors import CORSMiddleware
from wire_file import AsyncFileClient
from wiredb import AsyncChannel, Room, RoomManager

from .db import get_db
from .schemas import UserCreate, UserRead, UserUpdate
from .users import get_backend


class StoredRoom(Room):  # pragma: nocover
    def __init__(self, directory: str, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._directory = directory

    async def run(self, *args: Any, **kwargs: Any):
        await self.task_group.start(self.connect_to_file)
        await super().run(*args, **kwargs)

    async def connect_to_file(self, *, task_status: TaskStatus[None]) -> None:
        async with AsyncFileClient(
            doc=self.doc,
            path=f"{Path(self._directory) / self.id.lstrip('/')}.y",
        ):
            task_status.started()
            await sleep_forever()


class CocatApp:  # pragma: nocover
    def __init__(self, update_dir: str, db_path: str = "./test.db") -> None:
        db = get_db(db_path)
        backend = get_backend(db)

        @asynccontextmanager
        async def lifespan(app: FastAPI):
            async with RoomManager(
                partial(StoredRoom, update_dir)
            ) as self.room_manager:
                await db.create_db_and_tables()
                yield

        root_path = os.environ.get("COCAT_PROXY_PREFIX", "")
        if root_path:
            if not root_path.startswith("/"):
                root_path = "/" + root_path
            if root_path.endswith("/"):
                root_path = root_path[:-1]

        self.app = app = FastAPI(lifespan=lifespan, root_path=root_path)

        app.add_middleware(
            CORSMiddleware,
            allow_origins=[],
            allow_methods=["*"],
            allow_headers=["*"],
        )

        current_superuser = backend.fastapi_users.current_user(
            active=True, superuser=True
        )

        app.include_router(
            backend.fastapi_users.get_auth_router(backend.auth_backend),
            prefix="/auth/jwt",
            tags=["auth"],
        )
        app.include_router(
            backend.fastapi_users.get_register_router(UserRead, UserCreate),
            prefix="/auth",
            tags=["auth"],
            dependencies=[Depends(current_superuser)],
        )
        app.include_router(
            backend.fastapi_users.get_reset_password_router(),
            prefix="/auth",
            tags=["auth"],
        )
        app.include_router(
            backend.fastapi_users.get_verify_router(UserRead),
            prefix="/auth",
            tags=["auth"],
        )
        app.include_router(
            backend.fastapi_users.get_users_router(UserRead, UserUpdate),
            prefix="/users",
            tags=["users"],
        )

        @app.websocket("/room/{room_id}")
        async def connect_room(
            room_id: str,
            websocket=Depends(backend.websocket_auth),
        ):
            if websocket is None:
                return

            await websocket.accept()
            ywebsocket = YWebSocket(websocket, room_id)
            room = await self.room_manager.get_room(ywebsocket.id)
            await room.serve(ywebsocket)


class YWebSocket(AsyncChannel):  # pragma: nocover
    def __init__(self, websocket: WebSocket, path: str) -> None:
        self._websocket = websocket
        self._path = path

    @property
    def id(self) -> str:
        return self._path

    async def __anext__(self):
        try:
            return await self._websocket.receive_bytes()
        except WebSocketDisconnect:
            raise StopAsyncIteration()

    async def send(self, message: bytes) -> None:
        await self._websocket.send_bytes(message)

    async def receive(self) -> bytes:
        return await self._websocket.receive_bytes()



from .api import create_catalogue as create_catalogue
from .api import create_event as create_event
from .api import get_catalogue as get_catalogue
from .api import get_event as get_event
from .api import log_in as log_in
from .api import log_out as log_out
from .api import refresh as refresh
from .api import save as save
from .api import set_config as set_config
from .catalogue import Catalogue as Catalogue
from .cli import add_user_to_room as add_user_to_room
from .cli import create_user as create_user
from .cli import get_user as get_user
from .cli import remove_user_from_room as remove_user_from_room
from .cli import serve as serve
from .db import DB as DB
from .event import Event as Event
from .models import CatalogueModel as CatalogueModel
from .models import EventModel as EventModel
from .votable import export_votable_file as export_votable_file
from .votable import export_votable_str as export_votable_str
from .votable import import_votable_file as import_votable_file
from .votable import import_votable_str as import_votable_str



from collections.abc import AsyncGenerator, Callable
from dataclasses import dataclass
from typing import Awaitable

from anyio import Lock, Path
from fastapi import Depends
from fastapi_users.db import SQLAlchemyBaseUserTableUUID, SQLAlchemyUserDatabase
from fastapi_users_db_sqlalchemy.access_token import (
    SQLAlchemyBaseAccessTokenTableUUID,
)
from sqlalchemy import JSON, Column
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.orm import DeclarativeBase


class Base(DeclarativeBase):
    pass


class User(SQLAlchemyBaseUserTableUUID, Base):
    rooms = Column(JSON, default=[], nullable=False)


class AccessToken(SQLAlchemyBaseAccessTokenTableUUID, Base):
    pass


@dataclass
class DB:
    async_session_maker: async_sessionmaker[AsyncSession]
    create_db_and_tables: Callable[[], Awaitable]
    get_async_session: Callable[[], AsyncGenerator[AsyncSession, None]]
    get_user_db: Callable[[AsyncSession], AsyncGenerator[SQLAlchemyUserDatabase, None]]


def get_db(path: str) -> DB:
    database_url = f"sqlite+aiosqlite:///{path}"
    engine = create_async_engine(database_url)
    async_session_maker = async_sessionmaker(engine, expire_on_commit=False)
    lock = Lock()

    async def create_db_and_tables():
        async with lock:
            if await Path(path).exists():
                return

            async with engine.begin() as conn:  # pragma: nocover
                await conn.run_sync(Base.metadata.create_all)

    async def get_async_session() -> AsyncGenerator[AsyncSession, None]:
        async with async_session_maker() as session:
            yield session

    async def get_user_db(
        session: AsyncSession = Depends(get_async_session),
    ) -> AsyncGenerator[SQLAlchemyUserDatabase]:
        yield SQLAlchemyUserDatabase(session, User)

    return DB(
        async_session_maker=async_session_maker,
        create_db_and_tables=create_db_and_tables,
        get_async_session=get_async_session,
        get_user_db=get_user_db,
    )



import json
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from datetime import datetime
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Any
from uuid import uuid4

from .catalogue import Catalogue
from .db import DB, Event

if TYPE_CHECKING:
    from astropy.io.votable.tree import Field as VOField  # type: ignore[import-untyped]
    from astropy.io.votable.tree import (  # type: ignore[import-untyped]
        Table,
        VOTableFile,
    )


@dataclass
class _VOTableCocatField:
    python_type: type
    attr: dict[str, str]
    convert_vot: Callable[[Any], Any]
    convert_cocat: Callable[[Any], Any]
    cocat_name: str | None = None

    def name(self, field: "VOField") -> str:
        if self.cocat_name:
            return self.cocat_name
        return field.name

    def match(self, field: "VOField") -> bool:
        for k, v in self.attr.items():
            if field.__getattribute__(k) != v:
                return False
        return True

    def make_vot_field(self, table: "Table", name: str) -> "VOField":
        from astropy.io.votable.tree import Field as VOField

        if "name" not in self.attr:
            return VOField(table, name=name, **self.attr)
        return VOField(table, **self.attr)


class _VOTableCocatFieldSpecialDateTime(_VOTableCocatField):
    def __init__(self, attrs: dict[str, str], cocat_name: str | None = None) -> None:
        attrs.update(
            {"datatype": "char", "xtype": "dateTime", "utype": "", "arraysize": "*"}
        )
        super().__init__(datetime, attrs, datetime.isoformat, str, cocat_name)


def _vo_table_field_from(arg: type | str) -> _VOTableCocatField:
    vtf: _VOTableCocatField | None = None
    if isinstance(arg, str):
        for vtf in VOTABLE_COCAT_FIELDS:
            if vtf.cocat_name == arg:
                break
    else:
        for vtf in VOTABLE_COCAT_FIELDS:
            if vtf.cocat_name is None and vtf.python_type is arg:
                break

    assert vtf is not None
    return vtf


VOTABLE_COCAT_FIELDS = [
    _VOTableCocatFieldSpecialDateTime(
        {"name": "Start Time", "ID": "TimeIntervalStart", "ucd": "time.start"}, "start"
    ),
    _VOTableCocatFieldSpecialDateTime(
        {"name": "Stop Time", "ID": "TimeIntervalStop", "ucd": "time.end"}, "stop"
    ),
    _VOTableCocatFieldSpecialDateTime({}),
    _VOTableCocatField(
        int,
        {"datatype": "long"},
        lambda x: 0 if x is None else x,
        lambda x: None if x == 0 else x,
        "rating",
    ),
    _VOTableCocatField(int, {"datatype": "long"}, int, int),
    _VOTableCocatField(float, {"datatype": "double"}, float, float),
    _VOTableCocatField(bool, {"datatype": "boolean"}, bool, bool),
    _VOTableCocatField(
        list,
        {"datatype": "char", "arraysize": "*", "utype": "json"},
        json.dumps,
        json.loads,
    ),
    _VOTableCocatField(
        list,
        {"datatype": "char", "arraysize": "*", "name": "products"},
        json.dumps,
        json.loads,
        "products",
    ),
    _VOTableCocatField(
        list,
        {"datatype": "char", "arraysize": "*", "name": "tags"},
        json.dumps,
        json.loads,
        "tags",
    ),
    _VOTableCocatField(
        str, {"datatype": "char", "arraysize": "*"}, str, str
    ),  # last item, catch all strings
]

ATTRIBUTES = [
    ("start", _vo_table_field_from("start")),
    ("stop", _vo_table_field_from("stop")),
    ("author", _vo_table_field_from(str)),
    ("uuid", _vo_table_field_from(str)),
    ("tags", _vo_table_field_from(list)),
    ("products", _vo_table_field_from(list)),
    ("rating", _vo_table_field_from("rating")),
]

STANDARD_FIELDS = [v[0] for v in ATTRIBUTES]


def export_votable(catalogues: Sequence[Catalogue] | Catalogue) -> "VOTableFile":
    from astropy.io.votable.tree import Resource, TableElement, VOTableFile

    votable = VOTableFile()

    catalogue_list = (
        [catalogues] if isinstance(catalogues, Catalogue) else list(catalogues)
    )

    if len(catalogue_list) == 1:
        catalogue = catalogue_list[0]
        votable.description = f"Contact:{catalogue.author};Name:{catalogue.name}"

    resource = Resource()
    votable.resources.append(resource)

    attributes = list(ATTRIBUTES)

    all_attributes = set(
        attribute
        for catalogue in catalogue_list
        for attribute in catalogue.attributes.keys()
    )
    common_attributes = set.intersection(
        *[set(catalogue.attributes) for catalogue in catalogue_list]
    )

    for catalogue in catalogue_list:
        table = TableElement(votable, name=catalogue.name.replace(" ", "_"))
        resource.tables.append(table)

        all_attributes = set(
            attribute
            for event in catalogue.events
            for attribute in event.attributes.keys()
        )
        common_attributes = set.intersection(
            *[set(event.attributes) for event in catalogue.events]
        )

        # check that all attributes are present in every event
        if all_attributes != common_attributes:
            raise ValueError(
                "Export VOTable: not all attributes are present in all events "
                + f"{tuple(sorted(all_attributes - common_attributes))}"
            )

        for attr in sorted(all_attributes):
            # check that the type of all attribute values is identical
            attrs_value_types = list(
                set(type(event.attributes[attr]) for event in catalogue.events)
            )
            if len(attrs_value_types) != 1:
                raise ValueError(
                    "Export VOTable: not all value types are "
                    + f"identical for all events for attribute {attr}"
                )

            attributes.append((attr, _vo_table_field_from(attrs_value_types[0])))

        table.fields.extend(
            [vtf.make_vot_field(votable, name) for name, vtf in attributes]
        )

        table.create_arrays(len(catalogue.events))
        for i, event in enumerate(catalogue.events):
            values = []
            for k, vtf in attributes:
                if k in STANDARD_FIELDS:
                    v = getattr(event, k)
                else:
                    v = event.attributes.get(k, "")
                if isinstance(v, set):
                    v = list(v)
                values.append(vtf.convert_vot(v))
            table.array[i] = tuple(values)

    return votable


def import_votable(
    votable: "VOTableFile", db: DB, table_name: str | None = None
) -> None:
    author = "VOTable Import"
    table_name = table_name or f"Imported Catalogue from {datetime.now()}"

    if votable.description:
        for line in str(votable.description).split(";"):
            line = line.strip()
            values = line.split(":", 1)
            if len(values) != 2:
                continue  # pragma: nocover
            property, value = values
            value = value.strip()
            if property == "Contact":
                author = value
            elif property == "Name":
                table_name = value

    db_dict: dict[str, list[dict[str, Any]]] = {
        "catalogues": [],
        "events": [],
    }

    for i, table in enumerate(votable.iter_tables()):
        required_field_names: list[str] = ["Start Time", "Stop Time"]
        fields_vs_index: dict[tuple[int, str], _VOTableCocatField] = {}

        for j, field in enumerate(table.fields):
            if field.name in required_field_names:
                required_field_names.remove(field.name)

            # match field-event-signature to get converters/name
            for vtf in VOTABLE_COCAT_FIELDS:
                if vtf.match(field):
                    fields_vs_index[(j, vtf.name(field))] = vtf
                    break
            else:  # pragma: nocover
                raise ValueError(
                    f"VOTable import: cannot import field: {field.ID}, {field.name}, {field.datatype},"
                    + f" {field.xtype}"
                )

        if len(votable.resources[0].tables) == 1:
            this_name = table_name
        else:  # pragma: nocover
            this_name = f"{table_name}_{i}"

        if len(required_field_names) > 0:  # pragma: nocover
            raise ValueError(
                f"VOTable import: required fields are missing for table {this_name}"
            )

        catalogue: dict[str, Any] = {
            "name": this_name,
            "author": author,
            "events": [],
        }

        has_author_field = any(f[1] == "author" for f in fields_vs_index.keys())
        has_uuid_field = any(f[1] == "uuid" for f in fields_vs_index.keys())

        for el in table.array:
            event: dict[str, Any] = {"attributes": {}}
            if not has_author_field:  # pragma: nocover
                event["author"] = author

            if not has_uuid_field:  # pragma: nocover
                event["uuid"] = str(uuid4())

            for (index, name), vtf in fields_vs_index.items():
                if name in STANDARD_FIELDS:
                    event[name] = vtf.convert_cocat(el[index])
                else:
                    event["attributes"][name] = vtf.convert_cocat(el[index])

            if not any(event["uuid"] == e["uuid"] for e in db_dict["events"]):
                db_dict["events"].append(event)

            catalogue["events"].append(event["uuid"])  # type: ignore

        db_dict["catalogues"].append(catalogue)

    with db.transaction():
        for _event in db_dict["events"]:
            db.create_event(**_event)
        for _catalogue in db_dict["catalogues"]:
            events = _catalogue.pop("events", [])
            cat = db.create_catalogue(**_catalogue)
            cat.add_events([Event._from_uuid(uuid, db) for uuid in events])


def import_votable_file(
    file_path: str | Path, db: DB, table_name: str | None = None
) -> None:
    """
    Imports a VOTable file into a database.

    Args:
        file_path: The VOTable file path.
        db: The database into which to import the VOTable.
    """
    from astropy.io.votable import parse  # type: ignore[import-untyped]

    import_votable(parse(file_path), db, table_name=table_name)


def import_votable_str(xml_content: str, db: DB, table_name: str | None = None) -> None:
    """
    Imports a VOTable XML string into a database.

    Args:
        xml_content: The VOTable content as an XML string.
        db: The database into which to import the VOTable.
    """
    from astropy.io.votable import parse  # type: ignore[import-untyped]

    import_votable(parse(BytesIO(xml_content.encode())), db, table_name=table_name)


def export_votable_file(
    catalogues: Sequence[Catalogue] | Catalogue, file_path: str | Path
) -> None:
    """
    Exports catalogues to a VOTable file.

    Args:
        catalogues: The catalogue(s) to export.
        file_path: The path to the exported file.
    """
    with open(file_path, "wb") as f:
        export_votable(catalogues).to_xml(f)


def export_votable_str(catalogues: Sequence[Catalogue] | Catalogue) -> str:
    """
    Exports catalogues as a VOTable XML string.

    Args:
        catalogues: The catalogue(s) to export.

    Returns:
        The VOTable as an XML string.
    """
    content = BytesIO()
    export_votable(catalogues).to_xml(content)
    return content.getvalue().decode()



import uuid
from dataclasses import dataclass
from typing import Annotated, Any

from fastapi import Cookie, Depends, Path, Request, WebSocket, status
from fastapi_users import BaseUserManager, FastAPIUsers, UUIDIDMixin, models
from fastapi_users.authentication import (
    AuthenticationBackend,
    CookieTransport,
    JWTStrategy,
)
from fastapi_users.db import SQLAlchemyUserDatabase

from .db import User

SECRET = str(uuid.uuid4())


class UserManager(UUIDIDMixin, BaseUserManager[User, uuid.UUID]):
    reset_password_token_secret = SECRET
    verification_token_secret = SECRET

    async def on_after_register(self, user: User, request: Request | None = None):
        print(f"User {user.id} has registered.")

    async def on_after_forgot_password(
        self, user: User, token: str, request: Request | None = None
    ):
        print(
            f"User {user.id} has forgot their password. Reset token: {token}"
        )  # pragma: nocover

    async def on_after_request_verify(
        self, user: User, token: str, request: Request | None = None
    ):
        print(
            f"Verification requested for user {user.id}. Verification token: {token}"
        )  # pragma: nocover


@dataclass
class Backend:
    auth_backend: Any
    current_active_user: Any
    fastapi_users: Any
    get_jwt_strategy: Any
    get_user_manager: Any
    websocket_auth: Any


def get_backend(db) -> Backend:
    async def get_user_manager(
        user_db: SQLAlchemyUserDatabase = Depends(db.get_user_db),
    ):
        yield UserManager(user_db)

    async def websocket_auth(
        websocket: WebSocket,
        room_id: str = Path(...),
        fastapiusersauth: Annotated[str | None, Cookie()] = None,
        user_manager: BaseUserManager[models.UP, models.ID] = Depends(get_user_manager),
    ) -> WebSocket | None:  # pragma: nocover
        accept_websocket = False
        if fastapiusersauth is not None:
            user = await get_jwt_strategy().read_token(fastapiusersauth, user_manager)  # type: ignore[func-returns-value,arg-type]
            if user and room_id in user.rooms:
                accept_websocket = True
        if accept_websocket:
            return websocket

        await websocket.close(code=status.WS_1008_POLICY_VIOLATION)
        return None

    def get_jwt_strategy() -> JWTStrategy[models.UP, models.ID]:  # pragma: nocover
        return JWTStrategy(secret=SECRET, lifetime_seconds=3600)

    auth_backend: AuthenticationBackend = AuthenticationBackend(
        name="cookie",
        transport=CookieTransport(),
        get_strategy=get_jwt_strategy,
    )

    fastapi_users = FastAPIUsers[User, uuid.UUID](get_user_manager, [auth_backend])

    current_active_user = fastapi_users.current_user(active=True)

    return Backend(
        auth_backend=auth_backend,
        current_active_user=current_active_user,
        fastapi_users=fastapi_users,
        get_jwt_strategy=get_jwt_strategy,
        get_user_manager=get_user_manager,
        websocket_auth=websocket_auth,
    )



import contextlib
import traceback

from anycorn import Config
from anycorn import serve as anycorn_serve
from anyio import Event, run
from cyclopts import App
from fastapi_users.exceptions import UserAlreadyExists

from .app.app import CocatApp
from .app.db import get_db
from .app.schemas import UserCreate, UserUpdate
from .app.users import get_backend

app = App()


@app.command
def serve(
    *,
    host: str = "localhost",
    port: int = 8000,
    update_dir: str = "",
    db_path: str = "./test.db",
):
    """
    Launch a server.

    Args:
        host: The server host name
        port: The server port number
        update_dir: The path to the directory where the room updates are saved
        db_path: The path to the user database
    """
    run(_serve, host, port, update_dir, db_path)  # pragma: nocover


@app.command
def create_user(
    *,
    email: str,
    password: str,
    is_superuser: bool = False,
    db_path: str = "./test.db",
):
    """
    Create a new user in the database.

    Args:
        email: The user e-mail
        password: The user password
        is_superuser: Whether the user is a superuser
        db_path: The path to the user database
    """
    return run(_create_user, email, password, is_superuser, db_path)


def get_user(
    *,
    email: str,
    db_path: str = "./test.db",
):
    """
    Get a user from the database.

    Args:
        email: The user e-mail
        db_path: The path to the user database
    """
    return run(_get_user, email, db_path)


@app.command
def add_user_to_room(
    *,
    email: str,
    room_id: str,
    db_path: str = "./test.db",
):
    """
    Add a user to a room.

    Args:
        email: The user e-mail
        room_id: The room ID to add the user to
        db_path: The path to the user database
    """
    run(_add_user_to_room, email, room_id, db_path)


@app.command
def remove_user_from_room(
    *,
    email: str,
    room_id: str,
    db_path: str = "./test.db",
):
    """
    Remove a user from a room.

    Args:
        email: The user e-mail
        room_id: The room ID to remove the user from
        db_path: The path to the user database
    """
    run(_remove_user_from_room, email, room_id, db_path)


async def _serve(
    host: str, port: int, update_dir: str, db_path: str
):  # pragma: nocover
    config = Config()
    config.bind = [f"{host}:{port}"]
    shutdown_event = Event()
    tb = None
    try:
        cocat_app = CocatApp(update_dir, db_path)
        await anycorn_serve(
            cocat_app.app,  # type: ignore[arg-type]
            config,
            shutdown_trigger=shutdown_event.wait,
            mode="asgi",
        )
    except Exception:
        tb = traceback.format_exc()
    finally:
        if tb is not None:
            print(tb)
        shutdown_event.set()


async def _create_user(email: str, password: str, is_superuser: bool, db_path: str):
    db = get_db(db_path)
    await db.create_db_and_tables()
    backend = get_backend(db)
    get_async_session_context = contextlib.asynccontextmanager(db.get_async_session)
    get_user_db_context = contextlib.asynccontextmanager(db.get_user_db)
    get_user_manager_context = contextlib.asynccontextmanager(backend.get_user_manager)

    try:
        async with get_async_session_context() as session:
            async with get_user_db_context(session) as user_db:
                async with get_user_manager_context(user_db) as user_manager:
                    user = await user_manager.create(
                        UserCreate(
                            email=email, password=password, is_superuser=is_superuser
                        )
                    )
                    print(f"User created {user}")
                    return user
    except UserAlreadyExists:
        print(f"User {email} already exists")
        raise


async def _get_user(email: str, db_path: str):
    db = get_db(db_path)
    backend = get_backend(db)
    get_async_session_context = contextlib.asynccontextmanager(db.get_async_session)
    get_user_db_context = contextlib.asynccontextmanager(db.get_user_db)
    get_user_manager_context = contextlib.asynccontextmanager(backend.get_user_manager)

    async with get_async_session_context() as session:
        async with get_user_db_context(session) as user_db:
            async with get_user_manager_context(user_db) as user_manager:
                return await user_manager.get_by_email(email)


async def _add_user_to_room(email: str, room_id: str, db_path: str):
    db = get_db(db_path)
    backend = get_backend(db)
    get_async_session_context = contextlib.asynccontextmanager(db.get_async_session)
    get_user_db_context = contextlib.asynccontextmanager(db.get_user_db)
    get_user_manager_context = contextlib.asynccontextmanager(backend.get_user_manager)

    async with get_async_session_context() as session:
        async with get_user_db_context(session) as user_db:
            async with get_user_manager_context(user_db) as user_manager:
                user = await user_manager.get_by_email(email)
                user_update = UserUpdate(
                    email=user.email,
                    rooms=list(set(user.rooms) | set([room_id])),
                )
                await user_manager.update(user_update, user, False)


async def _remove_user_from_room(email: str, room_id: str, db_path: str):
    db = get_db(db_path)
    backend = get_backend(db)
    get_async_session_context = contextlib.asynccontextmanager(db.get_async_session)
    get_user_db_context = contextlib.asynccontextmanager(db.get_user_db)
    get_user_manager_context = contextlib.asynccontextmanager(backend.get_user_manager)

    async with get_async_session_context() as session:
        async with get_user_db_context(session) as user_db:
            async with get_user_manager_context(user_db) as user_manager:
                user = await user_manager.get_by_email(email)
                user_update = UserUpdate(
                    email=user.email,
                    rooms=list(set(user.rooms) - set([room_id])),
                )
                await user_manager.update(user_update, user, False)



import uuid

from fastapi_users import schemas
from pydantic import BaseModel


class CocatUser(BaseModel):
    rooms: list[str] = []


class UserRead(schemas.BaseUser[uuid.UUID], CocatUser):
    pass


class UserCreate(schemas.BaseUserCreate, CocatUser):
    pass


class UserUpdate(schemas.BaseUserUpdate, CocatUser):
    pass
